"""
Script that checks column integrated heat input of prescribed polar heating
"""
from glob import glob
import numpy as np
import xarray as xr
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as cm
import cartopy.crs as ccrs
from scipy import integrate
from aostools.constants import *

def open_heat_static(folder, filename):
    file = '/home/links/rm811/Isca/input/' + folder + '/' + filename + '.nc'
    ds = xr.open_dataset(file)
    lat = ds.lat
    lon = ds.lon
    p = ds.pfull
    heat = ds.variables[filename].transpose('lat','lon','pfull')
    return lat, p, lon, heat

def open_heat_evolution(filename):
    folder = '/disco/share/rm811/isca_data/PK_e0v4z13_' + filename
    files = sorted(glob(folder+'/run*/*.nc'))
    files_subset = files[:12]
    ds = xr.open_mfdataset(files_subset)
    t = ds.time.data
    heat = ds.local_heating
    return t, heat

def diff(heat, var):
    return heat.differentiate(var, edge_order=2)

def integral(heat, ticks, ax):
    return integrate.trapezoid(y = heat, x = ticks, axis = ax)

def time_integrate(heat, t):
    dTdt = diff(heat, 'time')
    return integral(dTdt, t, 0)

def surf_integrate(heat, lon, p):
    n = len(lon)
    if n == 1:
        heat = heat.mean('lon')
        s = integral(heat, p, 1)
        surf = s
        for j in range(127):
            surf = np.vstack((surf, s))
        surf = surf.transpose() 
    elif n > 1:
        surf = integral(heat, p, 2)
    return surf

def ratio(x0, x1):
    return x1/x0

def integration_levels(folders, filenames, level):
    # set-up the 3D grid first
    file = '/home/links/rm811/Isca/input/dimensions_3d.nc'
    ds = xr.open_dataset(file, decode_times=False)
    lat = ds.lat
    coslat = np.cos(np.deg2rad(lat.data))
    lon = ds.lon
    p = ds.pfull

    n = len(filenames)
    int1 = []
    for i in range(n):
       heat_lon, heat = open_heat_static(folders[i], filenames[i])[2:]
       int1.append(surf_integrate(heat, heat_lon, p))

    int2 = []
    for i in range(n):
        int2.append(integral(int1[i], lon, 1))

    if level == 'lat': 
        finished = int2
    
    elif level == 'full':
        int3 = []
        for i in range(n):
            int3_sub = []
            for j in range(len(int2[i])):
                int3_sub.append((int2[i][j]*coslat[j]))
            int3.append(np.sum(int3_sub)/np.sum(coslat))
        finished = int3
        print(finished)

    return n, lat, finished

def plot_vslat(folders, filenames, labels, colors):
    n, lat, int = integration_levels(folders, filenames, 'lat')
    fig, ax = plt.subplots(figsize=(6,6))
    for i in range(n):
        ax.plot(lat, int[i], color=colors[i], label=labels[i])
    ax.set_xlim(0, max(lat))
    ax.set_xlabel(r'latitude ($\degree$N)', fontsize='x-large')
    ax.set_ylim(bottom=0)
    ax.set_ylabel(r'integrated heat (W m$^{-2}$)', fontsize='x-large')
    plt.legend(fontsize='x-large')
    ax.tick_params(axis='both', labelsize = 'x-large', which='both', direction='in')
    plt.savefig('heatcheck1_polar.pdf', bbox_inches = 'tight')
    return plt.close()

def plot_vsdefault(folders, filenames, labels, colors):
    n, lat, int = integration_levels(folders, filenames, 'full')
    ratios = []
    for i in range(n):
        ratios.append(ratio(int[0], int[i]))
    print(ratios)
    fig, ax = plt.subplots(figsize=(8,6))
    ax.bar(labels[1:], ratios[1:], color=colors[1:])
    ax.set_xticklabels(labels[1:], rotation=30, ha='right')
    ax.axhline(ratios[0], color=colors[0], linewidth=1, linestyle='--')
    ax.text(0, ratios[0]+0.02, labels[0], color=colors[0], fontsize='x-large')
    ax.set_ylim(bottom=0)
    ax.set_ylabel(r'integrated heat ratio', fontsize='x-large')
    ax.tick_params(axis='both', labelsize = 'x-large', which='both', direction='in')
    plt.savefig('heatcheck2_polar.pdf', bbox_inches = 'tight')
    return plt.close()

def plot_multiheat():
    basis = 'PK_e0v4z13_'
    perturb = 'q6m2y45l800u200'
    T_ctrl = xr.open_dataset('/disco/share/rm811/processed/'+basis+perturb+'_Ttz.nc', decode_times=False).temp[0]
    heat_names = ['w15a4p800f800g50', 'w15a4p400f800g50', 'w30a4p800f800g50_s', 'q6m2y45l800u200']
    heats = []
    for heat in heat_names:
        print(heat)
        if heat == 'q6m2y45l800u200':
            folder = 'asymmetry'
            file = '/home/links/rm811/Isca/input/' + folder + '/' + heat + '.nc'
            ds = xr.open_dataset(file)
            heats.append(ds.sel(lon=180, method='nearest').variables[heat])
        else:
            folder = 'polar_heating'
            file = '/home/links/rm811/Isca/input/' + folder + '/' + heat + '.nc'
            ds = xr.open_dataset(file)
            heats.append(ds.mean('lon').variables[heat])


    lat = ds.lat
    p = ds.pfull
    h = 6/86400
    inc1 = 0.5e-5
    h_range1 = np.arange(inc1/2, h+inc1, inc1)
    inc2 = 1e-5
    h_range2 = np.arange(inc2/2, h+inc2, inc2)

    # Plot
    plt.figure(figsize=(7.5,7))
    plt.contour(T_ctrl.lat, T_ctrl.pfull, T_ctrl, colors='gainsboro', levels=np.arange(160, 330, 10), linewidths=1.5)
    cs = plt.contourf(lat, p, heats[0], cmap='Reds', levels=h_range1)
    cb = plt.colorbar(cs, location='right')
    cb.set_label(label=r'Heating (K s$^{-1}$)', size='xx-large')
    cb.ax.tick_params(labelsize='xx-large')
    plt.contour(lat, p, heats[1], colors='g', levels=h_range1, linewidths=1.5, alpha=0.15)
    plt.contour(lat, p, heats[2], colors='g', levels=h_range1, linewidths=1.5, alpha=0.15)
    plt.contour(lat, p, heats[3], colors='#4D0099', levels=h_range2, linewidths=1.5, alpha=0.25)
    plt.xlim(0, max(lat))
    plt.xticks([0, 20, 40, 60, 80], ['0', '20', '40', '60', '80'])
    plt.xlabel(r'Latitude ($\degree$N)', fontsize='xx-large')
    plt.ylim(max(p), 100)
    plt.yscale('log')
    plt.ylabel('Pressure (hPa)', fontsize='xx-large')
    plt.tick_params(axis='both', labelsize = 'xx-large', which='both', direction='in')
    plt.savefig('heatings.pdf', bbox_inches = 'tight')
    return plt.close()

folders = ['polar_heating', 'polar_heating', 'polar_heating', 'asymmetry', 'polar_heating']
filenames = ['w15a4p800f800g50', 'w30a4p800f800g50_scaled', 'w15a4p400f800g50', 'a4x75y180w5v30p800_s', 'w15a8p800f800g50']
labels = ['default', r'$\phi_{w} = 30\degree$', r'$p_{top} = 400$ hPa', 'off-pole', r'$A = 8$ K day$^{-1}$']
colors = ['k', '#B30000', '#FF9900', '#FFCC00', '#00B300', '#0099CC', '#4D0099', '#CC0080']

#plot_vslat(folders, filenames, labels, colors)
#plot_vsdefault(folders, filenames, labels, colors)
plot_multiheat()